"""Flag provides a boolean descriptor."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.desc import AbstractDescriptor
from worktoy.parse import maybe


class Flag(AbstractDescriptor):
  """Flag provides a boolean descriptor."""

  __default_value__ = None

  def __init__(self, defVal: bool = None) -> None:
    AbstractDescriptor.__init__(self)
    self.__default_value__ = maybe(defVal, False)

  def __instance_get__(self,
                       instance: object,
                       owner: type,
                       **kwargs) -> object:
    if instance is None:
      return self
    pvtName = self._getPrivateName()
    if getattr(instance, pvtName, None) is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      setattr(instance, pvtName, self.__default_value__)
      return self.__instance_get__(instance, owner, _recursion=True)
    return True if getattr(instance, pvtName) else False

  def __set__(self, instance: object, value: bool) -> None:
    """Setter-function implementation"""
    pvtName = self._getPrivateName()
    setattr(instance, pvtName, value)



"""The 'worktoy.ezdata' class provides the EZData class!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._field_tags import BeginFields, EndFields
from ._data_namespace import DataNamespace
from ._data_metaclass import DataMetaclass, EZData



"""DelayedDescriptor subclasses the AbstractDescriptor class and adds to
it the deferred object creation allowing instantiation of the inner class
to occur on an instance by instance basis.

Please note the nomenclature hierarchy:
  owner class > owner instance > dedicated instance of inner class
Since the name 'instance' firmly belongs to the instance of the owner class,
the name 'inner object' or similar is used to refer to the instance of the
inner class. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from worktoy.desc import AbstractDescriptor


class DelayedDescriptor(AbstractDescriptor):
  """DelayedDescriptor subclasses the AbstractDescriptor class and adds to
  it the deferred object creation allowing instantiation of the inner class
  to occur on an instance by instance basis.

  Please note the nomenclature hierarchy:
    owner class > owner instance > dedicated instance of inner class
  Since the name 'instance' firmly belongs to the instance of the owner
  class,
  the name 'inner object' or similar is used to refer to the instance of the
  inner class. """

  @abstractmethod
  def _createInnerObject(self, instance: object) -> object:
    """Creates an instance of the inner class. Please note that the
    subclass implementation will receive only the instance for whom to
    create the inner object. Subclasses must be able to create the inner
    object from the information provided up to this point. """

  def __instance_get__(self,
                       instance: object,
                       owner: type,
                       **kwargs) -> object:
    """The __instance_get__ method is called when the descriptor is accessed
    via the owning instance. If the instance already own an inner object,
    the inner object is returned. Otherwise, an inner object is created
    and assigned to the instance after which the method recursively calls
    itself to return the inner object. If for whatever reason the creation
    of the inner object fails, this is caught as a RecursionError."""
    pvtName = self._getPrivateName()
    if getattr(instance, pvtName, None) is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      innerObject = self._createInnerObject(instance)
      setattr(instance, pvtName, innerObject)
      return self.__instance_get__(instance, owner, _recursion=True)
    innerObject = getattr(instance, pvtName, None)
    if innerObject is None:
      e = """Failed to create inner object!"""
      raise AttributeError(e)
    return innerObject



"""The 'worktoy.desc' implements the descriptor protocol with lazy
instantiation. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._owner_instance import Owner, Instance
from ._attri_class import AttriClass
from ._empty_field import EmptyField
from ._abstract_descriptor import AbstractDescriptor
from ._flag import Flag
from ._delayed_descriptor import DelayedDescriptor
from ._typed_descriptor import TypedDescriptor
from ._attri_box import AttriBox



"""The 'worktoy.text' package contains modules for working with text data."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._mono_space import monoSpace
from ._join_words import joinWords
from ._string_list import stringList
from ._type_msg import typeMsg
from ._word_wrap import wordWrap



"""The 'maybeType' function receives a type followed by any number of
positional arguments. The function returns the first argument that is an
instance of the given type. The first argument may be a type, a tuple or
list of types, a string or a tuple or list of strings. When strings are
used (not recommended), the function matches the '__name__' attribute of
the types of the arguments.
"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from icecream import ic


def maybeType(*args) -> Any:
  """The 'maybeType' function receives a type followed by any number of
  positional arguments. The function returns the first argument that is an
  instance of the given type. The first argument may be a type, a tuple or
  list of types, a string or a tuple or list of strings. When strings are
  used (not recommended), the function matches the '__name__' attribute of
  the types of the arguments. """

  if not args:
    return None

  if isinstance(args[0], type):
    return _maybeType(args[0], *args[1:])

  if isinstance(args[0], str):
    return _maybeTypeName(args[0], *args[1:])

  if isinstance(args[0], (tuple, list)):
    if len(args[0]) == 0:
      return None
    for type_ in args[0]:
      val = None
      if isinstance(type_, type):
        val = _maybeType(type_, *args[1:])
        if val is not None:
          return val
      if isinstance(type_, str):
        val = _maybeTypeName(type_, *args[1:])
      if val is not None:
        return val


def _maybeTypeName(typeName: str, *args) -> Any:
  """This private function uses the '__name__' attribute of the types of the
  arguments to match the given type name."""
  for arg in args:
    if arg.__class__.__name__ == typeName:
      return arg


def _maybeType(type_: type, *args) -> Any:
  """This private function matches the given type."""
  for arg in args:
    if isinstance(arg, type_):
      return arg



"""The 'yolo' function receives any number of callables and runs them."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os
import sys
import time
from typing import Callable


def yolo(*args: Callable) -> None:
  """The 'yolo' function receives any number of callables and runs them."""
  tic = time.perf_counter_ns()
  print('Running python script located at: \n%s' % sys.argv[0])
  print('Started at: %s' % time.ctime())
  print(77 * '-')
  retCode = 0
  for callMeMaybe in args:
    print('\nRunning: %s\n' % callMeMaybe.__name__)
    try:
      retCode = callMeMaybe()
    except BaseException as exception:
      exceptionTypeName = exception.__class__.__name__
      exceptionMessage = str(exception)
      print('ENCOUNTERED!\n  %s: %s' % (exceptionTypeName, exceptionMessage))
      tb = exception.__traceback__
      while True:
        fileName = tb.tb_frame.f_code.co_filename
        if fileName == __file__:
          tb = tb.tb_next
          if tb is None:
            break
          continue
        lineNumber = tb.tb_lineno
        print("""In file: '%s', at line: %d""" % (fileName, lineNumber))
        with open(fileName, 'r') as file:
          data = file.readlines()
        for i in range(lineNumber - 3, lineNumber + 3):
          if 0 < i < len(data):
            line = data[i].replace(os.linesep, '')
            if i - lineNumber + 1:
              print('  %03d:    %s   ' % (i, line))
            else:
              print('  %03d: >> %s << ' % (i, line))
        if getattr(tb, 'tb_next', None):
          tb = tb.tb_next
        else:
          break

      retCode = -1
  retCode = 0 if retCode is None else retCode
  print(77 * '-')
  print('Return Code: %s' % retCode)
  toc = int(time.perf_counter_ns() - tic)
  n = 0
  while toc > 1e03:
    toc *= 1e-03
    n += 1
  name = ['nano', 'micro', 'milli', ''][int(n)]
  print('Runtime: %d %s-seconds' % (int(toc), name))



"""TypedDescriptor subclasses DelayedDescriptor and adds to it the inner
class. Instances of this class are desc that return inner objects
belonging to instances of the owning class. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from worktoy.desc import DelayedDescriptor, AttriClass
from worktoy.text import typeMsg


class TypedDescriptor(DelayedDescriptor):
  """TypedDescriptor subclasses DelayedDescriptor and adds to it the inner
  class. Instances of this class are desc that return inner objects
  belonging to instances of the owning class. """

  __inner_class__ = None

  def getInnerClass(self, **kwargs) -> type:
    """Returns the inner class. """
    if self.__inner_class__ is None:
      e = """The inner class has not been assigned. """
      raise AttributeError(e)
    if isinstance(self.__inner_class__, type):
      if hasattr(self.__inner_class__, '__ready_box__'):
        return self.__inner_class__
      if kwargs.get('_raw', False):
        return self.__inner_class__
      name = self.__inner_class__.__name__
      bases = (self.__inner_class__,)
      namespace = {**AttriClass.__dict__, }
      return type(name, bases, namespace)
    e = typeMsg('__inner_class__', self.__inner_class__, type)
    raise TypeError(e)

  def _setInnerClass(self, innerClass: type) -> None:
    """Sets the inner class. """
    if self.__inner_class__ is not None:
      e = """The inner class has already been assigned. """
      raise AttributeError(e)
    if not isinstance(innerClass, type):
      e = typeMsg('innerClass', innerClass, type)
      raise TypeError(e)
    self.__inner_class__ = innerClass

  @abstractmethod
  def _createInnerObject(self, instance: object) -> object:
    """Creates an instance of the inner class. """



"""AttriClass provides a minimal class that is ready for boxing. If a
custom class is intended to be used in AttriBox, it may subclass
AttriClass which will provide the necessary methods and attributes. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import typeMsg


class AttriClass:
  """AttriClass provides a minimal class that is ready for boxing. If a
  custom class is intended to be used in AttriBox, it may subclass
  AttriClass which will provide the necessary methods and attributes. """

  __ready_box__ = True
  __outer_box__ = None
  __owning_instance__ = None
  __field_owner__ = None
  __field_name__ = None

  def getFieldOwner(self) -> type:
    """Getter-function for the field owner. """
    return self.__field_owner__

  def setFieldOwner(self, owner: type) -> None:
    """Setter-function for the field owner. """
    if self.__field_owner__ is not None:
      e = """The field owner has already been assigned!"""
      raise AttributeError(e)
    if isinstance(owner, type):
      self.__field_owner__ = owner
    else:
      e = typeMsg('owner', owner, type)
      raise TypeError(e)

  def getFieldName(self) -> str:
    """Getter-function for the field name. """
    return self.__field_name__

  def setFieldName(self, name: str) -> None:
    """Setter-function for the field name. """
    if self.__field_name__ is not None:
      e = """The field name has already been assigned!"""
      raise AttributeError(e)
    if isinstance(name, str):
      self.__field_name__ = name
    else:
      e = typeMsg('name', name, str)
      raise TypeError(e)

  def getOwningInstance(self) -> object:
    """Getter-function for the owning instance. """
    return self.__owning_instance__

  def setOwningInstance(self, instance: object) -> None:
    """Setter-function for the owning instance. """
    if self.__owning_instance__ is not None:
      e = """The owning instance has already been assigned!"""
      raise AttributeError(e)
    self.__owning_instance__ = instance

  def getOuterBox(self, ) -> object:
    """Getter-function for the outer box. """
    return self.__outer_box__

  def setOuterBox(self, box: object) -> None:
    """Setter-function for the outer box. """
    if self.__outer_box__ is not None:
      e = """The outer box has already been assigned!"""
      raise AttributeError(e)
    self.__outer_box__ = box



"""This file provides some common type aliases used by 'worktoy.meta'. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import TypeAlias, Union, Tuple, TYPE_CHECKING, Callable

if TYPE_CHECKING:
  from worktoy.meta import AbstractNamespace
else:
  AbstractNamespace = object

Bases: TypeAlias = Union[type, Tuple[type, ...]]
Namespace: TypeAlias = Union[dict, AbstractNamespace]
Space: TypeAlias = Namespace
FunctionDecorator: TypeAlias = Callable[[Callable], Callable]
ClassDecorator: TypeAlias = Callable[[type], type]
Decorator: TypeAlias = Union[FunctionDecorator, ClassDecorator]



"""WorkToy provides a collection of tools."""
#  AGPL-3.0 license
#  Copyright (c) 2023-2024 Asger Jon Vistisen
from __future__ import annotations

import worktoy.parse
import worktoy.meta
import worktoy.text
import worktoy.desc
import worktoy.keenum
import worktoy.ezdata
import worktoy.yolo



"""SingletonMetaclass provides a custom metaclass for creating singletons.
These are classes having only one instance, namely the class itself."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import AbstractMetaclass


class SingletonMetaclass(AbstractMetaclass):
  """SingletonMetaclass provides a custom metaclass for creating singletons.
  These are classes having only one instance, namely the class itself."""

  __singleton_instance__ = None

  def __call__(cls, *args, **kwargs) -> object:
    """The __call__ method is invoked when the class is called."""
    if cls.__singleton_instance__ is None:
      cls.__singleton_instance__ = super().__call__(*args, **kwargs)
    return cls.__singleton_instance__



"""KeeNumObject provides the base class for all KeeNum classes. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Self

from worktoy.desc import EmptyField
from worktoy.text import typeMsg


class KeeNumObject:
  """KeeNumObject is the base class for all KeeNum classes. """

  __keenum_name__ = None
  __private_value__ = None
  __public_value__ = None

  name = EmptyField()
  _pvtVal = EmptyField()
  value = EmptyField()

  @name.GET
  def _getName(self) -> object:
    """Getter-function for the name."""
    if self.__keenum_name__ is None:
      e = """The name has not been assigned!"""
      raise AttributeError(e)
    if isinstance(self.__keenum_name__, str):
      return self.__keenum_name__
    e = typeMsg('__keenum_name__', self.__keenum_name__, str)
    raise TypeError(e)

  @name.SET
  def _setName(self, name: str) -> None:
    """Setter-function for the name."""
    if self.__keenum_name__ is not None:
      e = """The name has already been assigned!"""
      raise AttributeError(e)
    if not isinstance(name, str):
      e = typeMsg('name', name, str)
      raise TypeError(e)
    self.__keenum_name__ = name

  @_pvtVal.GET
  def _getPvtVal(self) -> object:
    """Getter-function for the private value."""
    if self.__private_value__ is None:
      e = """The private value has not been assigned!"""
      raise AttributeError(e)
    if isinstance(self.__private_value__, int):
      return self.__private_value__
    e = typeMsg('__private_value__', self.__private_value__, int)
    raise TypeError(e)

  @_pvtVal.SET
  def _setPvtVal(self, privateValue: int) -> None:
    """Setter-function for the private value."""
    if self.__private_value__ is not None:
      e = """The private value has already been assigned!"""
      raise AttributeError(e)
    if not isinstance(privateValue, int):
      e = typeMsg('privateValue', privateValue, int)
      raise TypeError(e)
    self.__private_value__ = privateValue

  @value.GET
  def _getPubVal(self) -> object:
    """Getter-function for the public value."""
    return self.__public_value__

  @value.SET
  def _setPubVal(self, publicValue: object) -> None:
    """Setter-function for the public value."""
    if self.__public_value__ is not None:
      e = """The public value has already been assigned!"""
      raise AttributeError(e)
    self.__public_value__ = publicValue

  def __init__(self, *args, ) -> None:
    for arg in args:
      if isinstance(arg, KeeNumObject):
        self.__keenum_name__ = arg.name
        self.__private_value__ = arg._pvtVal
        self.__public_value__ = arg.value
        break

  def __init_subclass__(cls, **kwargs) -> None:
    """Why are we still here? Just to suffer? Or to raise errors?"""

  def __str__(self) -> str:
    """String representation"""
    return '%s.%s' % (self.__class__.__name__, self.name)

  def __int__(self) -> int:
    """Exposes the private value"""
    if isinstance(self.__private_value__, int):
      return self.__private_value__
    e = typeMsg('__private_value__', self.__private_value__, int)
    raise TypeError(e)

  def __add__(self, other: object) -> Self:
    """Implements addition"""
    if isinstance(other, int):
      return self.__class__(int(self) + other)
    if isinstance(other, KeeNumObject):
      return self + int(other)
    return NotImplemented

  def __sub__(self, other: object) -> Self:
    """Implements subtraction"""
    if isinstance(other, int):
      return self.__class__(int(self) - other)
    if isinstance(other, KeeNumObject):
      return self - int(other)
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Implementation of equal operator"""
    if isinstance(other, int):
      return False if int(self) - other else True
    if isinstance(other, KeeNumObject):
      return self == int(other)
    if isinstance(other, self.__class__):
      return self == int(other)
    return NotImplemented



"""AttriBox subclasses the TypedDescriptor class and incorporates
syntactic sugar for setting the inner class, and for the inner object
creation. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Callable, Never, Any

from icecream import ic

from worktoy.desc import TypedDescriptor, Instance, Owner
from worktoy.parse import maybe
from worktoy.text import typeMsg, monoSpace

if sys.version_info.minor < 11:
  from typing_extensions import Self
else:
  from typing import Self

ic.configureOutput(includeContext=True)

if TYPE_CHECKING:
  pass
else:
  AttriClass = object


class AttriBox(TypedDescriptor):
  """AttriBox subclasses the TypedDescriptor class and incorporates
  syntactic sugar for setting the inner class, and for the inner object
  creation. """

  __positional_args__ = None
  __keyword_args__ = None
  __get_callbacks__ = None
  __set_callbacks__ = None
  __del_callbacks__ = None

  def _getGetCallbacks(self) -> list[Callable]:
    """Getter-function for list of functions to be called on get."""
    if self.__get_callbacks__ is None:
      self.__get_callbacks__ = []
    return self.__get_callbacks__

  def _getSetCallbacks(self) -> list[Callable]:
    """Getter-function for list of functions to be called on set."""
    if self.__set_callbacks__ is None:
      self.__set_callbacks__ = []
    return self.__set_callbacks__

  def _getDelCallbacks(self) -> list[Callable]:
    """Getter-function for list of functions to be called on del."""
    if self.__del_callbacks__ is None:
      self.__del_callbacks__ = []
    return self.__del_callbacks__

  def notifyGet(self, callMeMaybe: Callable) -> Callable:
    """Adds given callable to list of callables to be notified on get."""
    self._getGetCallbacks().append(callMeMaybe)
    return callMeMaybe

  def notifySet(self, callMeMaybe: Callable) -> Callable:
    """Adds given callable to list of callables to be notified on set."""
    self._getSetCallbacks().append(callMeMaybe)
    return callMeMaybe

  def notifyDel(self, callMeMaybe: Callable) -> Callable:
    """Adds given callable to list of callables to be notified on del."""
    self._getDelCallbacks().append(callMeMaybe)
    return callMeMaybe

  def ONGET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for adding a function to the get callbacks."""
    return self.notifyGet(callMeMaybe)

  def ONSET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for adding a function to the set callbacks."""
    return self.notifySet(callMeMaybe)

  def ONDEL(self, callMeMaybe: Callable) -> Callable:
    """Decorator for adding a function to the del callbacks."""
    return self.notifyDel(callMeMaybe)

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the AttriBox instance. """
    TypedDescriptor.__init__(self, *args, **kwargs)
    if not kwargs.get('_root', False):
      e = """The AttriBox class should not be instantiated directly!"""
      raise TypeError(e)
    if not args:
      e = """The inner class must be provided. """
      raise TypeError(e)
    innerClass = args[0]
    if not isinstance(innerClass, type):
      e = typeMsg('innerClass', innerClass, type)
      raise TypeError(e)
    self._setInnerClass(innerClass)

  @classmethod
  def __class_getitem__(cls, innerClass: type) -> Self:
    """Syntactic sugar for setting the inner class. """
    return cls(innerClass, _root=True)

  def __call__(self, *args, **kwargs) -> Self:
    """Syntactic sugar for creating an instance of the inner class. """
    self.__positional_args__ = args
    self.__keyword_args__ = kwargs
    return self

  def _getArgs(self, instance: object) -> list:
    """Returns the arguments used to create the inner object. """
    out = []
    for arg in maybe(self.__positional_args__, []):
      if arg is Instance:
        out.append(instance)
      elif arg is Owner:
        out.append(self._getFieldOwner())
      else:
        out.append(arg)
    return out

  def _getKwargs(self, instance: object) -> dict:
    """Returns the keyword arguments used to create the inner object. """
    out = {}
    for (key, value) in maybe(self.__keyword_args__, {}).items():
      if value is Instance:
        out[key] = instance
      elif value is Owner:
        out[key] = self._getFieldOwner()
      else:
        out[key] = value
    return self.__keyword_args__

  def _createInnerObject(self, instance: object) -> object:
    """Creates an instance of the inner class. """
    innerClass = self.getInnerClass()
    args, kwargs = self._getArgs(instance), self._getKwargs(instance)
    innerObject = innerClass(*args, **kwargs)
    if TYPE_CHECKING:
      assert isinstance(innerObject, AttriClass)
    innerObject.setOuterBox(self)
    innerObject.setOwningInstance(instance)
    innerObject.setFieldOwner(self._getFieldOwner())
    innerObject.setFieldName(self._getFieldName())
    return innerObject

  def __str__(self, ) -> str:
    try:
      fieldName = self._getFieldName()
      ownerName = self._getFieldOwner().__name__
    except AttributeError as attributeError:
      if 'has not been assigned to a field' not in str(attributeError):
        raise attributeError
      ownerName = '(TBD)'
      fieldName = '(TBD)'
    innerName = self.getInnerClass().__name__
    return '%s.%s: %s' % (ownerName, fieldName, innerName)

  def __repr__(self, ) -> str:
    try:
      fieldName = self._getFieldName()
    except AttributeError as attributeError:
      if 'has not been assigned to a field' not in str(attributeError):
        raise attributeError
      fieldName = '(TBD)'
    innerName = self.getInnerClass().__name__
    args = [*self.__positional_args__, *self.__keyword_args__]
    args = ', '.join([str(arg) for arg in args])
    return '%s = AttriBox[%s](%s)' % (fieldName, innerName, args)

  @classmethod
  def _getOwnerListName(cls) -> str:
    """Returns the name at which the list of attribute instances of this
    type. Please note that this name is not unique to the owner as they
    are in separate scopes."""
    return '__boxes_%s__' % cls.__qualname__

  def __set_name__(self, owner: type, name: str) -> None:
    """Sets the name of the field. """
    ownerListName = self._getOwnerListName()
    if not getattr(owner, '__has_boxes__', False):
      setattr(owner, '__has_boxes__', True)

    TypedDescriptor.__set_name__(self, owner, name)
    existing = getattr(owner, ownerListName, [])
    if existing:
      return setattr(owner, ownerListName, [*existing, self])
    setattr(owner, ownerListName, [self, ])
    oldInitSub = getattr(owner, '__init_subclass__')

    def newInitSub(cls, *args, **kwargs) -> None:
      """Triggers the extra init"""
      oldInitSub(*args, **kwargs)
      self.applyBoxes(cls)

    setattr(owner, '__init_subclass__', classmethod(newInitSub))

  @classmethod
  def applyBoxes(cls, owner: type) -> None:
    """Applies the boxes to the owner class."""
    ownerListName = cls._getOwnerListName()
    boxes = getattr(owner, ownerListName, [])
    for box in boxes:
      if not isinstance(box, AttriBox):
        e = typeMsg('box', box, AttriBox)
        raise TypeError(e)
      boxName = box._getFieldName()
      setattr(cls, boxName, box)
      cls.__set_name__(box, owner, boxName)

  def __get__(self, instance: object, owner: type) -> Any:  # Footnote
    """The __get__ method is called when the descriptor is accessed via the
    owning instance. """
    value = TypedDescriptor.__get__(self, instance, owner)
    for callback in self._getGetCallbacks():
      callback(instance, value)
    return value

  def __set__(self, instance: object, value: object) -> None:
    """The __set__ method is called when the descriptor is assigned a value
    via the owning instance. """
    pvtName = self._getPrivateName()
    oldValue = getattr(instance, pvtName, None)
    setattr(instance, pvtName, value)
    for callback in self._getSetCallbacks():
      callback(instance, oldValue, value)

  def __delete__(self, instance: object, ) -> Never:
    """Deleter-function not yet implemented!"""
    e = """Tried deleting the '%s' attribute from instance of class '%s', 
    but this deleter-function is not yet implemented!"""
    msg = e % (self._getFieldName(), instance.__class__.__name__)
    raise NotImplementedError(monoSpace(msg))

#  Footnote: 'Any' or 'object'? Type hinting to 'Any' tells the type
#  checker to ignore this return value. Presently, that is the best the
#  static type checker can manage. However, type hinting to 'object' tells
#  the type checker that the return value cannot be assumed to belong to
#  any type. Thus, the type checker will indicate a problem if an object
#  is assumed belonging to a specific type when it was returned with type
#  hinting to 'object'. This is generally preferable, but in this
#  particular case, the type checker is not able to infer the type of the
#  objects created with AttriBox. For example:
#   ______________________________________________________________________
#  | class Foo:                                                           |
#  |   bar = AttriBox[int]()                                              |
#  |                                                                      |
#  |   def someMethod(self, ) -> int:                                     |
#  |     return self.bar  # Type checker will indicate a problem here!    |
#   ¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨
#  The type checker is not able to infer that the AttriBox instance is in
#  fact an 'int'. However, this is an exception. In most cases,
#  type hinting to 'object' is preferable. The 'correct' way to handle this
#  would require the use of a stub file. It is possible to implement in
#  the AttriBox class functionality that creates a stub file at the first
#  runtime that sets the type hint to the type in the bracket. For the
#  above example, the stub file would be:
#   _____________________________________________________________________
#  | class Foo:                                                          |
#  |   bar: int                                                          |
#   ¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨
#  Presently, having to use 'Any' instead of 'object' is a problem.
#  Implementing dynamic stub file creation will be very powerful,
#  but requires a significant effort. In the future, when more problems
#  are identified that will benefit dynamic stub file creation, this will
#  be one of the problems solved.
#  -- Asger Jon Vistisen, July 18, 2024 --
#  UPDATE, same day --
#  Writing a function that dynamically creates the code required by the
#  stub file and saving it was not as much work as the above note might
#  have indicated. Nevertheless, a bit more work is needed to add a unit
#  test. So far, the implementation is not able to look for an existing
#  stub file, but a future improvement will enable different scripts to
#  add rather than just overwrite.
#  UPDATE, again same day --
#  Implementation of the dynamically created stub file is now working. All
#  that remains is to add the unit tests. This does provide somewhat of a
#  novelty, as the unit test must test if a file is created. To ensure
#  that a unittest is not confused by an existing class from previous runs,
#  we use the 'setUpClass' and 'tearDownClass' methods to create and delete
#  the stub file.
#  UPDATE, again same day --
#  The unit tests are now implemented. This should allow the AttriBox
#  __get__ method to type hint to 'object', as the stub file will be
#  created automatically. What remains is to include methods in the
#  created stub files and to support multiple files in the same stub file.
#  UPDATE, again same day --
#  The stub files generated are missing the import statements.
#  UPDATE, again same day --
#  The stub files now include import statements. These are even formatted
#  correctly to support nested packages. Still missing is to include the
#  methods and to handle the case of multiple classes in the same file.
#  UPDATE, again same day --
#  Moved stub file generation to dedicated function.



"""KeeNumSpace provides the namespace for the KeeNuMeta class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.keenum import KeeNumObject
from worktoy.meta import AbstractNamespace
from worktoy.parse import maybe
from worktoy.text import typeMsg


def auto(*args) -> KeeNumObject:
  """Creates an entry in the KeeNum class."""
  keeNumObject = KeeNumObject()
  if args:
    if len(args) == 1:
      keeNumObject.value = args[0]
    else:
      keeNumObject.value = (*args,)
  return keeNumObject


class KeeNumSpace(AbstractNamespace):
  """KeeNumSpace is a Zeroton object indicating the class owning the
  descriptor instance."""

  __keenum_entries__ = None

  def getKeeNumEntries(self) -> list[KeeNumObject]:
    """Getter-function for the KeeNum entries."""
    out = maybe(self.__keenum_entries__, [])
    if isinstance(out, list):
      for item in out:
        if not isinstance(item, KeeNumObject):
          e = typeMsg('item', item, KeeNumObject)
          raise TypeError(e)
      return out
    e = typeMsg('__keenum_entries__', out, list)
    raise TypeError(e)

  def __explicit_set__(self, key: str, value: object) -> None:
    """The __explicit_set__ method is invoked by __setitem__."""
    if key == '__init__':
      e = """The __init__ method is reserved for KeeNum classes!"""
      raise AttributeError(e)
    if isinstance(value, KeeNumObject):
      key = key.upper()
      entries = self.getKeeNumEntries()
      value._pvtVal = len(entries)
      value.name = key
      if value.value is None:
        value.value = key
      self.__keenum_entries__ = [*entries, value]

  def compile(self) -> dict:
    """The compile method returns the namespace without the KeeNumObjects."""
    out = {}
    entries = self.getKeeNumEntries()
    for key, value in self.items():
      if value in entries:
        continue
      out[key] = value
    return out



"""DataNamespace provides the namespace object for the DataMetaclass."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Callable

from icecream import ic

from worktoy.desc import AttriBox
from worktoy.ezdata import EndFields, BeginFields
from worktoy.meta import AbstractNamespace
from worktoy.parse import maybe
from worktoy.text import typeMsg

ic.configureOutput(includeContext=True)


class DataNamespace(AbstractNamespace):
  """DataNamespace provides the namespace object for the DataMetaclass."""

  __data_fields__ = None
  __field_section__ = False

  def _initFactory(self, ) -> Callable:
    """Factory function for the __init__ method"""

    dataFields = self.getDataFields()

    def newInit(this, *args, **kwargs) -> None:
      """The newInit method is the __init__ method for the class."""
      for (arg, (key, field)) in zip(args, dataFields.items()):
        setattr(this, key, arg)
      for (key, field) in dataFields.items():
        if key in kwargs:
          val = kwargs[key]
          setattr(this, key, val)

    return newInit

  def getDataFields(self) -> dict[str, AttriBox]:
    """The getDataFields method returns the data fields."""
    dataFields = maybe(self.__data_fields__, {})
    if isinstance(dataFields, dict):
      for (key, val) in dataFields.items():
        if not isinstance(key, str):
          e = typeMsg('key', key, str)
          raise TypeError(e)
        if not isinstance(val, AttriBox):
          e = typeMsg('val', val, AttriBox)
          raise TypeError(e)
      return dataFields
    e = typeMsg('dataFields', dataFields, dict)
    raise TypeError(e)

  def _appendDataField(self, key: str, field: object) -> None:
    """Appends the AttriBox instance at the given key."""
    existing = self.getDataFields()
    if key in existing:
      e = """Name conflict at name: '%s'!""" % key
      raise NameError(e)
    self.__data_fields__ = {**existing, key: field}

  def __explicit_set__(self, key: str, value: object) -> None:
    """The __setitem__ hook collects AttriBox instances. """
    if key == '__init__':
      e = """Reimplementing __init__ is not allowed for EZData classes!"""
      raise AttributeError(e)
    if self.__field_section__:
      self._appendDataField(key, value)

  def compile(self, ) -> dict[str, object]:
    """The compile method returns the class namespace."""
    out = {}
    for (key, val) in dict.items(self):
      out[key] = val
    out['__init__'] = self._initFactory()
    return out

  def __getitem__(self, key: str) -> object:
    """The __getitem__ method returns the value at the given key."""
    try:
      return AbstractNamespace.__getitem__(self, key)
    except KeyError as keyError:
      if BeginFields == key:
        if self.__field_section__:
          e = """Field section already started!"""
          raise ValueError(e)
        self.__field_section__ = True
      if EndFields == key:
        if not self.__field_section__:
          e = """Field section not started!"""
          raise ValueError(e)
        self.__field_section__ = False
      raise keyError
